/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

#include "prelude.h"

// renvoie l'indice de début d'une occurrence du motif `m` dans le texte `t`,
// ou -1 s'il n'y a pas d'occurrence
int naive_search(char *m, char *t) {
  int lm = strlen(m), lt = strlen(t);
  for (int i = 0; i <= lt - lm; i++) {
    int j = 0;
    while (j < lm) {
      if (m[j] != t[i+j]) break;
      j++;
    }
    if (j == lm) return i;
  }
  return -1;
}

// compte le nombre d'occurrences du motif `m` dans le texte `t`
int count_occurrences(char *m, char *t) {
  int lm = strlen(m), lt = strlen(t);
  int count = 0;
  for (int i = 0; i <= lt - lm; i++) {
    int j = 0;
    while (j < lm) {
      if (m[j] != t[i+j]) break;
      j++;
    }
    if (j == lm) { count++; }
  }
  return count;
}

// renvoie la longueur de la plus longue séquence ininterrompue
// d'un même caractère
int longest_repetition(char *t) {
  int lt = strlen(t);
  int i = 0, r = 0;
  while (i < lt) {
    int k = 1;
    while (i + k < lt && t[i + k] == t[i]) { k++; }
    if (k > r) { r = k; }
    i += k;
  }
  return r;
}

// `t` est-il un palindrome ?
bool palindrome(char *t) {
  int lo = 0, hi = strlen(t) - 1;
  while (lo < hi) {
    if (t[lo] != t[hi]) return false;
    lo++; hi--;
  }
  return true;
}

// renvoie la longueur du plus long palindrome de `t`
int longest_palindrome(char *t) {
  int a = 0, b = 0;
  int l_max = 0, n = strlen(t);
  while (b < n) {
    int r = 0;
    while (a - r >= 0 && b + r < n && t[a - r] == t[b + r]) r++;
    int l = b - a + 2 * r - 1;
    if (l > l_max) { l_max = l; }
    if (a == b) { b++; }
    else { a++; }
  }
  return l_max;
}

int manacher_longest_palindrome(char *t) {
  int c = 0;
  int l_max = 0, r = 0;
  int n = strlen(t);
  int *radii = calloc(2 * n + 1, sizeof(int));
  while (c <  2 * n + 1) {
    int a = c / 2, b = (c + 1) / 2;
    while (a - r >= 0 && b + r < n && t[a - r] == t[b + r]) r++;
    int l = b - a + 2 * r - 1;
    if (l > l_max) { l_max = l; }
    radii[c] = r;
    int old_c = c, old_r = r;
    c++;
    r = 0;
    while (c <= old_c + old_r) {
      int c_mirror = old_c - (c - old_c);
      int max_known = old_c + old_r - c;
      if (radii[c_mirror] < max_known) {
        radii[c] = radii[c_mirror];
        c++;
      } else if (radii[c_mirror] > max_known) {
        radii[c] = max_known;
        c++;
      } else {
        r = max_known;
        break;
      }
    }
  }
  return l_max;
}
